/*
 ** Module   : Quoter.java
 ** Abstract : Java implementation of the Quoter utility
 **
 ** Copyright (c) 2012, Golden Code Development Corporation.
 ** ALL RIGHTS RESERVED. Use is subject to license terms.
 **
 **           Golden Code Development Corporation
 **                      CONFIDENTIAL
 **
 ** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
 ** 001 CS  20121102          Created initial version.

 */
package com.goldencode.p2j.util;

import java.io.*;
import java.util.*;

/**
 * A class implementation of the Progress QUOTER utility.
 * <p>
 * This class is used to parse the quoter utility parameters and execute the
 * equivalent logic of the Progress utility.
 */
public class Quoter
{
   /**
    * Private constructor to stop class instantiation
    */
   private Quoter()
   {
      
   }

   /**
    * The entry point method
    * 
    * @param    strings
    *           The parameters given to the quoter utility
    */
   public static void launch(String[] strings)
   {
      String inFile = null;
      String outFile = null;
      String delimiter = null;
      String colRanges = null;
      
       // (first parameter will be the utility name -> "quoter" and can be
       // ignored)
      for (int i = 1; i < strings.length; i++)
      {
         if (strings[i].toUpperCase().equals("-D"))
         {
            delimiter = strings[i + 1];
         }
         if (strings[i].toUpperCase().equals("-C"))
         {
            colRanges = strings[i + 1];
         }
         if (strings[i].startsWith(">"))
         {
            inFile = strings[i - 1];
            outFile = strings[i].substring(1, strings[i].length());
         }
      }
      parseLaunchQuoter(inFile, outFile, delimiter, colRanges);
   }

   /**
    * Utility method that parses the parameters then calls the main logic of
    * quoter.
    * 
    * @param    inFile
    *           The input file name
    * @param    outFile
    *           The output file name
    * @param    delimiter
    *           The delimiter if any
    * @param    colRange
    *           The column ranges to be parsed
    */
   private static void parseLaunchQuoter(String inFile,
                                         String outFile,
                                         String delimiter,
                                         String colRange)
   {
      File input = new File(inFile);
      File output = new File(outFile);
      String separator = null;
      List<ColumnRange> columnRanges = null;
      if (delimiter != null)
      {
         separator = delimiter;
      }
      
      BufferedReader readFile = null;
      BufferedWriter destFile = null;
      try
      {
         destFile = new BufferedWriter(new FileWriter(output));
      }
      catch (IOException e1)
      {
         return;
      }
      
      if (delimiter != null && colRange != null)
      {
         // We cannot have both column and delimiter options
         try
         {
            destFile.write("Incompatible options: -c and -d ");
            destFile.close();
         }
         catch (IOException e)
         {
            return;
         }
         return;
      }

      try
      {
         //get column ranges if any.
         columnRanges = getColumnRange(colRange, destFile);
      }
      catch (Exception e)
      {
         // something went wrong return
         return;
      }

      if (input.exists())
      {
         try
         {
            readFile = new BufferedReader(new FileReader(input));
         }
         catch (FileNotFoundException e)
         {
            // something happened with the input (deleted)
            if (destFile != null)
            {
               try
               {
                  destFile.write("Unable to read: " + inFile);
                  destFile.close();
               }
               catch (IOException e1)
               {
                  return;
               }
            }
         }

         try
         {
            quoterWorker(readFile,
                         destFile,
                         columnRanges,
                         separator);
         }
   
         catch (IOException e)
         {
            // something went wrong and closing resources failed retry and end it
            try
            {
               readFile.close();
               destFile.close();
            }
            catch (IOException e1)
            {
               return;
            }
         }

         catch (Exception e)
         {
            // something went wrong but resources were closed. End it!
            return;
         }
      }
      else
      {
         // provided input does not exist
         if (destFile != null)
         {
            try
            {
               destFile.write("Unable to read: " + inFile);
               destFile.close();
            }
            catch (IOException e)
            {
               return;
            }
         }
      }

      try
      {
         if (readFile != null)
         {
            readFile.close();
         }
         destFile.close();
      }
      catch (IOException e)
      {
         return;
      }
   }

   /**
    * Utility method that writes into the destination file the line of quoted
    * tokens (row of columns parsed from the input file).
    *
    * @param   tokens 
    *          The <code>array</code> of <code>String</code> values 
    *          representing a row of columns parsed from the input file
    * @param   writer  
    *          The <code>BufferedWriter</code> used to output the quoted
    *          characters.
    *                
    * @throws  IOException
    *          In case something went wrong in the writing process.This indicates
    *          that the main quoter method should close opened resources and
    *          terminate.
    */
   private static void writeDelimitedLine(String[] tokens, 
                                          BufferedWriter writer)
   throws IOException
   {
      for (int i = 0; i < tokens.length; i++)
      {
         if (i != 0)
         {
            writer.write(" ");
         }
         writer.append('"');
         writer.write(tokens[i]);
         writer.append('"');
      }
   }

   /**
    * Utility method that identifies the columns from an input <code>String</code>
    * line extracts and quotes them then outputs them to the destination file
    * 
    * @param   line 
    *          The <code>String</code> line representing a row from the input 
    *          file.
    * @param   columnRanges 
    *          The <code>List</code> of <code>ColumnRange</code> used to 
    *          identify the column from the input line.
    * @param   writer
    *          The <code>BufferedWriter</code> used to output the quoted
    *          characters.
    * 
    * @throws  IOException
    *          In case something went wrong in the writing process.This indicates
    *          that the main quoter method should close opened resources and
    *          terminate.
    */
   private static void writeColumnsLine(String             line,
                                    List<ColumnRange>  columnRanges,
                                    BufferedWriter     writer)
   throws IOException
   {
      boolean first = true;
      for (ColumnRange column : columnRanges)
      {
         String columnText = null;
         if (line.length() >= column.getEndColumn())
         {
            columnText = line.substring(column.getStartColumn(),
                                        column.getEndColumn());
         }
         else
         {
            if (column.getStartColumn() > line.length())
            {
               columnText = null;
            }
            else
            {
               columnText = line.substring(column.getStartColumn());
            }
         }

         if (first)
         {
            first = false;
         }
         else
         {
            writer.write(" ");
         }

         if (columnText != null)
         {
            writer.append('"');
            writer.write(columnText);
            writer.append('"');
         }
         else
         {
            // when column text is null Progres puts a "-" mark
            writer.write("-");
         }
      }
   }

   /**
    * Utility method that parses a <code>String</code> containing column range
    * and returns the <code>List</code> of <code>ColumnRange<code> that 
    * contains the character range from a line where the column can be found.
    *
    * @param   colRange
    *          The <code>String</code> representation of the column ranges.
    * @param   destFile
    *          The <code>BufferedWriter</code> used to output the quoted
    *          characters or any error messages in this case
    * 
    * @return  The parsed <code>List</code> of <code>ColumnRange</code>.
    * 
    * @throws  Exception
    *          In case there are invalid column ranges. This indicates that the 
    *          main quoter method should close opened resources and terminate.
    */
   private static List<ColumnRange> getColumnRange(String          colRange, 
                                                   BufferedWriter  destFile) 
   throws Exception
   {
      List<ColumnRange> columnRanges = null;
      if (colRange != null)
      {
         columnRanges = new ArrayList<ColumnRange>();

         String[] ranges = colRange.split(",");
         for (int i = 0; i < ranges.length; i++)
         {
            String[] columns = ranges[i].split("-");
            if (columns.length == 2)
            {
               int startCol = Integer.parseInt(columns[0]);
               int endCol = Integer.parseInt(columns[1]);
               // check for invalid collumn ranges
               if (startCol > endCol)
               {
                  if (destFile != null)
                  {
                     try
                     {  
                        // output error to destination file
                        destFile.write("Invalid range: " + ranges[i]);
                        destFile.close();
                        // Invalid range - Throw exception to mark end
                        throw new Exception();
                     }
                     catch (IOException e)
                     {
                        throw e;
                     }
                  }
               }
               else
               {
                  columnRanges.add(new ColumnRange(startCol, endCol));
               }
            }
         }
      }

      return columnRanges;
   }

   /**
    * Utility method that takes a line from the input file , quotes it and 
    * outputs it to the destination file.
    * 
    * @param   readFile
    *          The input <code>BufferedReader</code> file from which the lines
    *          will be quoted
    * @param   destFile
    *          The destination <code>BufferedWriter</code> file to which the quoted
    *          lines will be output.
    * 
    * @throws  Exception 
    *          In case resources were closed successfully.This will signal 
    *          further that the quoter utility should be stopped without 
    *          additional steps.
    * @throws  IOException
    *          In case there was a problem closing the resources.This will signal
    *          further that the quoter utility should be stopped but there were
    *          problems closing the resources and it might be best to try and 
    *          close them again.
    */
   private static void quoteSimpleLine(BufferedReader readFile,
                                       BufferedWriter destFile) 
   throws IOException, Exception
   {
      String line = null;
      boolean firstElement = true;
      try
      {
         while ((line = readFile.readLine()) != null)
         {
            if (firstElement)
            {
               
               firstElement = false;
            }
            else
            {
               destFile.newLine();
            }
 
            destFile.write('"' + line + '"');
         }
      }
      catch (IOException e)
      {
         closeResourcesPreExit(readFile, destFile);
      }
   }

   /**
    * Utility method that reads a file line by line, splits the line using a
    * separator and outputs the quoted tokens to the destination file.
    * 
    * @param   readFile
    *          The input <code>BufferedReader</code> file from which the lines
    *          will be read and split into tokens.
    * @param   destFile
    *          The destination <code>BufferedWriter</code> file to which the 
    *          quoted lines will be output.
    * @param   separator
    *          The <code>String</code> character that is used to split the information
    *          from a line into columns.
    *         
    * @throws  Exception 
    *          In case resources were closed successfully.This will signal 
    *          further that the quoter utility should be stopped without 
    *          additional steps.
    * @throws  IOException
    *          In case there was a problem closing the resources.This will signal
    *          further that the quoter utility should be stopped but there were
    *          problems closing the resources and it might be best to try and 
    *          close them again.
    *  
    */
   private static void quoteDelimiter(BufferedReader     readFile,
                                      BufferedWriter     destFile,
                                      String             separator) 
   throws IOException, 
          Exception
   {
      String line = null;
      boolean firstElement = true;
      try
      {
         while ((line = readFile.readLine()) != null)
         {
            // Delimiter case
            String[] tokens = line.split(separator);
            if (firstElement)
            {
               firstElement = false;
            }
            else
            {
               destFile.newLine();
            }

            writeDelimitedLine(tokens, destFile);
         }
      }
      catch (IOException e)
      {
         closeResourcesPreExit(readFile, destFile);
      }
   }

   /**
    * Utility method that reads a file line by line, extracts the info using 
    * column ranges quoting the info and outputting it to a destination file.
    *
    * @param   readFile
    *          The input <code>BufferedReader</code> file from which the lines
    *          will be read and split into columns.
    * @param   destFile
    *          The destination <code>BufferedWriter</code> file to which the 
    *          quoted lines will be output.
    * @param   columnRanges
    *          The <code>ColumnRange</code> <code>List</code> containing the 
    *          column range definitions.
    * 
    * @throws  Exception 
    *          In case resources were closed successfully.This will signal 
    *          further that the quoter utility should be stopped without 
    *          additional steps.
    * @throws  IOException
    *          In case there was a problem closing the resources.This will signal
    *          further that the quoter utility should be stopped but there were
    *          problems closing the resources and it might be best to try and 
    *          close them again.
    */
   private static void quoteColumns(BufferedReader     readFile,
                                    BufferedWriter     destFile,
                                    List<ColumnRange>  columnRanges) 
   throws IOException, 
          Exception
   {
      String line = null;
      boolean firstElement = true;
      try
      {
         while ((line = readFile.readLine()) != null)
         {
            if (columnRanges.size() > 0)
            {
               if (firstElement)
               {
                  firstElement = false;
               }
               else
               {
                  destFile.newLine();
               }

               writeColumnsLine(line, columnRanges, destFile);
            }
         }
      }
      catch (IOException e)
      {
         closeResourcesPreExit(readFile, destFile);
      }
   }
   
   /**
    * Utility method that handles the logic of quoter utility.
    * 
    * @param   readFile
    *          The input <code>BufferedReader</code> file from which the info
    *          will be read.
    * @param   destFile
    *          The destination <code>BufferedWriter</code> file to which the 
    *          quoted lines will be output.
    * @param   columnRanges
    *          The <code>List</code> of <code>ColumnRange</code> used to 
    *          identify the column from the input line.
    * @param   delimiter
    *          The <code>String</code> character that is used to split the information
    *          from a line into columns.
    *  
    * @throws  IOException
    *          In case there was a problem closing the resources.This will signal
    *          further that the quoter utility should be stopped but there were
    *          problems closing the resources and it might be best to try and 
    *          close them again.
    * @throws  Exception 
    *          In case resources were closed successfully.This will signal 
    *          further that the quoter utility should be stopped without 
    *          additional steps.
    */
   private static void quoterWorker(BufferedReader     readFile,
                                    BufferedWriter     destFile,
                                    List<ColumnRange>  columnRanges,
                                    String             delimiter)
   throws IOException, Exception
   {
      if (readFile != null)
      {
         if (delimiter != null)
         {
            // delimiter case
            quoteDelimiter(readFile, destFile, delimiter);
         }
         else
         {
            if (columnRanges != null)
            { 
               // column range case
               quoteColumns(readFile, destFile, columnRanges);
            }
            else
            {
               // The simple case no delimiter no column range
               quoteSimpleLine(readFile, destFile);
            }
         }
      }
   }

   /**
    * Utility method that will close the two given resource (input and output)
    * and throw <code>Exception</code> if close was successful or
    * <code>IOException</code> if close was unsuccessful.
    * 
    * @param   input
    *          The input resource.
    * @param   output
    *          The output resource.
    *         
    * @throws  IOException
    *          In case there was a problem closing the resources.This will signal
    *          further that the quoter utility should be stopped but there were
    *          problems closing the resources and it might be best to try and 
    *          close them again.
    * @throws  Exception 
    *          In case resources were closed successfully.This will signal 
    *          further that the quoter utility should be stopped without 
    *          additional steps.
    */
   private static void closeResourcesPreExit(Closeable input, Closeable output) 
   throws IOException, 
          Exception
   {
      if (input != null && output != null)
      {
         try
         {
            input.close();
            output.close();
         }
         catch (IOException e)
         {
            throw e;
         }
      }

      throw new Exception();
   }
   
   /**
    * Utility class to handle column ranges
    */
   private static class ColumnRange
   {
      /**
       * This represents the starting point from where a column will  be
       * subtracted from a <code> String </code> line.
       */
      private int startColumn;
      /**
       * This represents the end point to where a column will be subtracted
       * from a <code> String </code> line.
       */
      private int endColumn;

      /**
       * Constructor with 2 parameters. Which will set the start and end index
       * defining a column. 
       * @param   startColumn
       *          The index point from where the column starts(position 
       *          starts from 1).
       * @param   endColumn
       *          The end index to where the column ends (inclusive).
       */
      public ColumnRange(int startColumn, int endColumn)
      {
         // Java starts from 0 not from 1 
         this.startColumn = startColumn - 1;
         // in Progress endColumn is inclusive so no need to substract 
         this.endColumn = endColumn;
      }
      
      /**
       * Returns the starting index of the column.
       * @return   The Java equivalent starting index of the column
       */
      public int getStartColumn()
      {
         return startColumn;
      }

      /**
       * Returns the end index of the column.
       * @return   The Java equivalent end index of the column.
       */
      public int getEndColumn()
      {
         return endColumn;
      }
   }
   
}